/* src/vm/jit/arm/md-abi.c - functions for arm ABI

   Copyright (C) 1996-2005 R. Grafl, A. Krall, C. Kruegel, C. Oates,
   R. Obermaisser, M. Platter, M. Probst, S. Ring, E. Steiner,
   C. Thalinger, D. Thuernbeck, P. Tomsich, C. Ullrich, J. Wenninger,
   Institut f. Computersprachen - TU Wien

   This file is part of CACAO.

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2, or (at
   your option) any later version.

   This program is distributed in the hope that it will be useful, but
   WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
   02111-1307, USA.

   Contact: cacao@complang.tuwien.ac.at

   Authors: Michael Starzinger

   Changes:

   $Id: md-abi.c,v 1.2 2005/07/09 01:54:58 michi Exp $

*/


#include "vm/jit/arm/types.h"
#include "vm/jit/arm/md-abi.h"

#include "vm/descriptor.h"
#include "vm/global.h"


/* md_param_alloc **************************************************************

   Allocate Arguments to Stackslots according the Calling Conventions

   --- in:
   md->paramcount:           Number of arguments for this method
   md->paramtypes[].type:    Argument types

   --- out:
   md->params[].inmemory:    Argument spilled on stack
   md->params[].regoff:      Stack offset or rd->arg[int|flt]regs index
   md->memuse:               Stackslots needed for argument spilling
   md->argintreguse:         max number of integer arguments used
   md->argfltreguse:         max number of float arguments used

*******************************************************************************/

void md_param_alloc(methoddesc *md)
{
	paramdesc *pd;
	s4         i;
	s4         reguse;
	s4         stacksize;

	/* set default values */
	reguse = 0;
	stacksize = 0;

	/* get params field of methoddesc */
	pd = md->params;

	for (i = 0; i < md->paramcount; i++, pd++) {
		switch (md->paramtypes[i].type) {
		case TYPE_INT:
		case TYPE_ADR:
		case TYPE_FLT:
			if (reguse < INT_ARG_CNT) {
				pd->inmemory = false;
				pd->regoff = reguse;
				reguse++;
			} else {
				pd->inmemory = true;
				pd->regoff = stacksize;
				stacksize++;
			}
			break;
		case TYPE_LNG:
		case TYPE_DBL:
			if (reguse+1 < INT_ARG_CNT) {
				pd->inmemory = false;
				pd->regoff = PACK_REGS(reguse, reguse+1);
				reguse += 2;
			} else if (reguse < INT_ARG_CNT) {
				printf("md_param_alloc(): we have a split here!\n");
				/*pd->inmemory = false;
				pd->regoff = reguse;
				reguse++;
				stacksize++;*/
				/* TODO: this is only a hack! */
				pd->inmemory = true;
				pd->regoff = stacksize;
				stacksize += 2;
				/* TODO: this hack prevents argument mixing */
				/* fill up empty argument register */
				reguse++;
			} else {
				pd->inmemory = true;
				pd->regoff = stacksize;
				stacksize += 2;
			}
			break;
		}
	}

	/* Since R0/R1 (==A0/A1) are used for passing return values, this         */
	/* argument register usage has to be regarded, too                        */
	if (md->returntype.type != TYPE_VOID) {
		if (!IS_2_WORD_TYPE(md->returntype.type)) {
			if (reguse < 1) reguse = 1;
		} else {
			if (reguse < 2) reguse = 2;
		}
	}

	/* fill register and stack usage */
	md->argintreguse = reguse;
	md->argfltreguse = 0; /* TODO: dont forget this */
	md->memuse = stacksize;
}


/* md_return_alloc *************************************************************

   Precolor the Java Stackelement containing the Return Value, if possible.

   --- in
   m:                       Methodinfo of current method
   return_type:             Return Type of the Method (TYPE_INT.. TYPE_ADR)
                            TYPE_VOID is not allowed!
   stackslot:               Java Stackslot to contain the Return Value

   --- out
   if precoloring was possible:
   stackslot->varkind       =ARGVAR
            ->varnum        =-1
            ->flags         =0
            ->regoff        =[REG_RESULT, (REG_RESULT2/REG_RESULT), REG_FRESULT]
   rd->arg[flt|int]reguse   set to a value according the register usage

*******************************************************************************/

void md_return_alloc(methodinfo *m, registerdata *rd, s4 return_type,
                                         stackptr stackslot) {
	/* In Leafmethods Local Vars holding parameters are precolored to their   */
	/* argument register -> so leafmethods with paramcount > 0 could already  */
	/* use  R0 == a00! */
	if (!m->isleafmethod || (m->paramcount == 0)) {
	/* Only precolor the stackslot, if it is not a SAVEDVAR <-> has not       */
	/* to survive method invokations                                          */
	if (!(stackslot->flags & SAVEDVAR)) {

		/*printf("md_return_alloc(): we have precoloring here (%s.%s)!\n", m->class->name->text, m->name->text);*/
		/* TODO: i think we cannot precolor floats and doubles */
		if ( IS_FLT_DBL_TYPE(return_type) )
		{
			printf("md_return_alloc(): we have a float/double here (%s)!\n", m->name->text);
			return;
		}

		stackslot->varkind = ARGVAR;
		stackslot->varnum = -1;
		stackslot->flags = 0;
		if (!IS_2_WORD_TYPE(return_type)) {
			if (rd->argintreguse < 1) rd->argintreguse = 1;
			stackslot->regoff = REG_RESULT;
		} else {
			if (rd->argintreguse < 2) rd->argintreguse = 2;
			stackslot->regoff = PACK_REGS(REG_RESULT, REG_RESULT2);
		}

	}
	}
}


/*
 * These are local overrides for various environment variables in Emacs.
 * Please do not remove this and leave it at the end of the file, where
 * Emacs will automagically detect them.
 * ---------------------------------------------------------------------
 * Local variables:
 * mode: c
 * indent-tabs-mode: t
 * c-basic-offset: 4
 * tab-width: 4
 * End:
 */
